# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""RPKI EE Certificate implementation - RFC6487."""

from __future__ import annotations

import logging
import typing

from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import padding

from . import base, ca, oid

if typing.TYPE_CHECKING:
    from ..sigobj.base import EncapsulatedContentType, SignedObject

log = logging.getLogger(__name__)

ECT = typing.TypeVar("ECT", bound="EncapsulatedContentType")


class EECertificate(base.BaseResourceCertificate, typing.Generic[ECT]):
    """RPKI EE Certificate - RFC6487."""

    def __init__(self, *,
                 signed_object: SignedObject[ECT],
                 **kwargs: typing.Any) -> None:
        """Initialise the EE Certificate."""
        self._signed_object = signed_object
        common_name = signed_object.econtent.signed_attrs_digest()
        super().__init__(common_name=common_name, **kwargs)

    @property
    def signed_object(self) -> SignedObject[ECT]:
        """Get the SignedObject that this certificate signs."""
        return self._signed_object

    @property
    def issuer_repo_path(self) -> str:
        """Get the filesystem path to the the issuer publication point."""
        return typing.cast(ca.CertificateAuthority, self.issuer).repo_path

    @property
    def mft_entry(self) -> typing.Optional[base.ManifestEntryInfo]:
        """Get an entry for inclusion in the issuer's manifest."""
        return (self.signed_object.file_name,
                self.signed_object.to_der())

    @property
    def sia(self) -> typing.Optional[x509.SubjectInformationAccess]:
        """Get the SubjectInformationAccess extension for the certificate."""
        sia_obj_uri = f"{self.base_uri}/" \
                      f"{self.issuer_repo_path}/" \
                      f"{self.signed_object.file_name}"
        sia = x509.SubjectInformationAccess([
            x509.AccessDescription(oid.SIA_OBJ_ACCESS_OID,
                                   x509.UniformResourceIdentifier(sia_obj_uri)),  # noqa: E501
        ])
        return sia

    def sign_object(self) -> bytes:
        """Construct a signature over the signedAttrs of the SignedObject."""
        message = self.signed_object.econtent.signed_attrs().to_der()
        signature = self.private_key.sign(data=message,
                                          padding=padding.PKCS1v15(),
                                          algorithm=self.hash_algorithm())
        return signature

    def publish(self, *, pub_path: str, **kwargs: typing.Any) -> None:
        """Publish the SignedObject artifact as a DER file in the PP."""
        self.signed_object.publish(pub_path=pub_path,
                                   uri_path=self.uri_path,
                                   repo_path=self.issuer_repo_path,
                                   **kwargs)



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Base RPKI Resource Certificate implementation - RFC6487."""

from __future__ import annotations

import datetime
import logging
import os
import typing
import urllib.parse

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from . import asn1, extensions, oid
from ..resources import AsResourcesInfo, IpResourcesInfo

if typing.TYPE_CHECKING:
    from .ca import CertificateAuthority

log = logging.getLogger(__name__)

ManifestEntryInfo = typing.Tuple[str, bytes]


class BaseResourceCertificate:
    """Base RPKI Resource Certificate class - RFC6487."""

    # rfc6487 section 4.3
    hash_algorithm = hashes.SHA256

    def __init__(self, *,  # noqa: R701
                 common_name: str,
                 days: int = 365,
                 issuer: typing.Optional[CertificateAuthority] = None,
                 ca: bool = False,
                 base_uri: str = "rsync://rpki.example.net/rpki",
                 ip_resources: typing.Optional[IpResourcesInfo] = None,
                 as_resources: typing.Optional[AsResourcesInfo] = None) -> None:  # noqa: E501
        """Initialise the Resource Certificate."""
        log.info(f"doing base initialisation of {self}")
        self._issuer = issuer
        self._base_uri = urllib.parse.urlparse(base_uri)

        builder = x509.CertificateBuilder()

        # rfc6487 section 4.2
        if self.issuer is None:
            serial_number = typing.cast("CertificateAuthority",
                                        self).next_serial_number
        else:
            serial_number = self.issuer.next_serial_number
        builder = builder.serial_number(serial_number)
        # rfc6487 section 4.5
        self._cn = common_name
        subject_name = x509.Name([x509.NameAttribute(x509.NameOID.COMMON_NAME,
                                                     self.subject_cn)])
        builder = builder.subject_name(subject_name)
        # rfc6487 section 4.4
        if self.issuer is None:
            issuer_name = subject_name
        else:
            issuer_name = self.issuer.cert.subject
        builder = builder.issuer_name(issuer_name)
        # rfc6487 section 4.6
        valid_from = datetime.datetime.utcnow()
        valid_to = valid_from + datetime.timedelta(days=days)
        builder = builder.not_valid_before(valid_from) \
                         .not_valid_after(valid_to)
        # rfc6487 sect 4.7 and rfc7935 section 3
        self._key = rsa.generate_private_key(public_exponent=65537,
                                             key_size=2048)
        builder = builder.public_key(self.public_key)
        # rfc6487 section 4.8.1
        if ca is True:
            basic_constraints = x509.BasicConstraints(ca=True,
                                                      path_length=None)
            builder = builder.add_extension(basic_constraints, critical=True)
        # rfc6487 section 4.8.2
        ski = x509.SubjectKeyIdentifier.from_public_key(self.public_key)
        self._ski_digest = ski.digest
        builder = builder.add_extension(ski, critical=False)
        # rfc6487 section 4.8.3
        if self.issuer is not None:
            aki = x509.AuthorityKeyIdentifier\
                      .from_issuer_public_key(self.issuer.public_key)
            builder = builder.add_extension(aki, critical=False)
        # rfc6487 section 4.8.4
        key_usage = x509.KeyUsage(digital_signature=ca is False,
                                  key_cert_sign=ca is True,
                                  crl_sign=ca is True,
                                  content_commitment=False,
                                  key_encipherment=False,
                                  data_encipherment=False,
                                  key_agreement=False,
                                  encipher_only=False,
                                  decipher_only=False)
        builder = builder.add_extension(key_usage, critical=True)
        # rfc6487 section 4.8.6
        if self.issuer is not None and self.issuer.crldp is not None:
            builder = builder.add_extension(self.issuer.crldp, critical=False)
        # rfc6487 section 4.8.7
        if self.issuer is not None and self.issuer.aia is not None:
            builder = builder.add_extension(self.issuer.aia, critical=False)
        # rfc6487 section 4.8.8
        if self.sia is not None:
            builder = builder.add_extension(self.sia, critical=False)
        # rfc6487 section 4.8.9
        if self.cps is not None:
            builder = builder.add_extension(self.cps, critical=True)
        # rfc6487 section 4.8.10
        if ip_resources is not None:
            ip_resources_ext = extensions.IpResources(ip_resources)
            builder = builder.add_extension(ip_resources_ext, critical=True)
        # rfc6487 section 4.8.11
        if as_resources is not None:
            as_resources_ext = extensions.AsResources(as_resources)
            builder = builder.add_extension(as_resources_ext, critical=True)

        self._cert_builder = builder

        if self.issuer is None:
            self._cert = typing.cast("CertificateAuthority", self).issue_cert()
        else:
            self._cert = self.issuer.issue_cert(self)

    @property
    def sia(self) -> typing.Optional[x509.SubjectInformationAccess]:
        """Construct the SIA extension."""
        raise NotImplementedError

    @property
    def cps(self) -> typing.Optional[x509.CertificatePolicies]:
        """Construct the CPS extension."""
        cps = x509.CertificatePolicies([
            x509.PolicyInformation(oid.RPKI_CERT_POLICY_OID,
                                   policy_qualifiers=None),
        ])
        return cps

    @property
    def mft_entry(self) -> typing.Optional[ManifestEntryInfo]:
        """Get an entry for inclusion in the issuer's manifest."""
        raise NotImplementedError

    def publish(self, *,
                pub_path: str,
                recursive: bool = True,
                **kwargs: typing.Any) -> None:
        """Publish artifact files in the PP."""
        raise NotImplementedError

    @property
    def private_key(self) -> rsa.RSAPrivateKey:
        """Get the private part of the RSA key pair."""
        return self._key

    @property
    def public_key(self) -> rsa.RSAPublicKey:
        """Get the public part of the RSA key pair."""
        return self._key.public_key()

    @property
    def cert_builder(self) -> x509.CertificateBuilder:
        """Get the certificate builder used to construct the certificate."""
        return self._cert_builder

    @property
    def base_uri(self) -> str:
        """Get the base URI of the RPKI publication service."""
        return self._base_uri.geturl()

    @property
    def uri_path(self) -> str:
        """Get the relative filesystem path equivalent of base_uri."""
        return os.path.join(self._base_uri.hostname or "",
                            *self._base_uri.path.rstrip("/").split("/"))

    @property
    def cert(self) -> x509.Certificate:
        """Get the underlying cryptography X.509 Certificate object."""
        return self._cert

    @property
    def cert_der(self) -> bytes:
        """Get cert DER-encoded."""
        return self.cert.public_bytes(serialization.Encoding.DER)

    @property
    def issuer(self) -> typing.Optional[CertificateAuthority]:
        """Get the issuing CertificateAuthority."""
        return self._issuer

    @property
    def subject_cn(self) -> str:
        """Get the common_name component of the subjectName."""
        return self._cn

    @property
    def ski_digest(self) -> bytes:
        """Get the message digest of the SKI extension."""
        return self._ski_digest

    @property
    def asn1_cert(self) -> asn1.Certificate:
        """Get an ASN.1 Certificate for the certificate."""
        log.info(f"Constructing ASN.1 Certificate from {self}")
        log.debug(f"Using DER bytes:\n{self.cert_der.hex()}")
        return asn1.Certificate.from_der(self.cert_der)

    @property
    def subject_public_key_info(self) -> asn1.SubjectPublicKeyInfo:
        """Get the subjectPublicKeyInfo for the certificate."""
        return self.asn1_cert.subject_public_key_info


ResourceCertificates = typing.Iterable[BaseResourceCertificate]
ResourceCertificateList = typing.List[BaseResourceCertificate]



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""rpki-conjure command implementation."""

from __future__ import annotations

import argparse
import importlib.metadata
import ipaddress
import logging
import os
import typing

from . import Args, BaseCommand, Return
from .helpers import as_id_or_range, ip_resource, roa_network

if typing.TYPE_CHECKING:
    from ..cert import CertificateAuthority

log = logging.getLogger(__name__)

DEFAULT_OUTPUT_DIR = os.path.join(os.curdir, "target", "demo")
PUB_SUB_DIR = "repo"
TAL_SUB_DIR = "tals"

DEFAULT_TA_AS_RESOURCES = [(0, 4294967295)]
DEFAULT_TA_IP_RESOURCES = [ipaddress.ip_network("0.0.0.0/0"),
                           ipaddress.ip_network("::0/0")]

DEFAULT_CA_AS_RESOURCES = [65000, (65010, 65019)]
DEFAULT_CA_IP_RESOURCES = [ipaddress.ip_network("10.0.0.0/8"),
                           (ipaddress.ip_address("192.168.0.0"),
                            ipaddress.ip_address("192.168.2.255")),
                           ipaddress.ip_network("2001:db8::/32")]

DEFAULT_GBR_FULLNAME = "Jane Doe"
DEFAULT_GBR_ORG = "Example Org"
DEFAULT_GBR_EMAIL = "jane@example.net"

META_PATH = "<path>"
META_AS = "<asn>"
META_AS_RANGE = f"{META_AS}[-{META_AS}]"
META_IP_PREFIX = "<prefix>/<length>"
META_IP_PREFIX_MAXLEN = f"{META_IP_PREFIX}[-maxlen]"
META_IP_RANGE = "<addr-lower>-<addr-upper>"
META_IP = f"{META_IP_PREFIX}|{META_IP_RANGE}"
META_NAME = "<name>"
META_ADDR = "<addr>"


class Conjure(BaseCommand):
    """Conjure a fully populated RPKI repository from thin air."""

    subcommand = "conjure"

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialise subcommand."""
        super().__init__(*args, **kwargs)
        log.info("trying to load plugins")
        self._plugins = list()
        entry_point_name = "rpkimancer.cli.conjure"
        entry_points = importlib.metadata.entry_points()
        for entry_point in entry_points.get(entry_point_name, []):
            cls = entry_point.load()
            if issubclass(cls, ConjurePlugin):
                plugin = cls(self.parser)
                self._plugins.append(plugin)

    def init_parser(self) -> None:
        """Set up command line argument parser."""
        self.parser.add_argument("--output-dir", "-o",
                                 default=DEFAULT_OUTPUT_DIR,
                                 metavar=META_PATH,
                                 help="Directory to write generated artifacts to "  # noqa: E501
                                      "(default: %(default)s)")
        self.parser.add_argument("--ta-as-resources",
                                 nargs="+", type=as_id_or_range,
                                 default=DEFAULT_TA_AS_RESOURCES,
                                 metavar=META_AS_RANGE,
                                 help="ASN(s) to include in TA certificate "
                                      "(default: %(default)s)")
        self.parser.add_argument("--ta-ip-resources",
                                 nargs="+", type=ip_resource,
                                 default=DEFAULT_TA_IP_RESOURCES,
                                 metavar=META_IP,
                                 help="IP addresses to include in TA certificate "  # noqa: E501
                                      "(default: %(default)s)")
        self.parser.add_argument("--ca-as-resources",
                                 nargs="+", type=as_id_or_range,
                                 default=DEFAULT_CA_AS_RESOURCES,
                                 metavar=META_AS_RANGE,
                                 help="ASN(s) to include in suboridinate CA certificate "  # noqa: E501
                                      "(default: %(default)s)")
        self.parser.add_argument("--ca-ip-resources",
                                 nargs="+", type=ip_resource,
                                 default=DEFAULT_CA_IP_RESOURCES,
                                 metavar=META_IP,
                                 help="IP addresses to include in suboridinate CA certificate "  # noqa: E501
                                      "(default: %(default)s)")
        self.parser.add_argument("--roa-asid",
                                 type=int,
                                 default=DEFAULT_CA_AS_RESOURCES[0],
                                 metavar=META_AS,
                                 help="ASN to include in ROA asID "
                                      "(default: %(default)s)")
        self.parser.add_argument("--roa-networks",
                                 nargs="+", type=roa_network,
                                 default=[(net, None)
                                          for net in DEFAULT_CA_IP_RESOURCES
                                          if isinstance(net, (ipaddress.IPv4Network,  # noqa: E501
                                                              ipaddress.IPv6Network))],  # noqa: E501
                                 metavar=META_IP_PREFIX_MAXLEN,
                                 help="IP prefixes to include in ROA "
                                      "(default: %(default)s)")
        self.parser.add_argument("--gbr-full-name",
                                 default=DEFAULT_GBR_FULLNAME,
                                 metavar=META_NAME,
                                 help="Full name to include in GBR "
                                      "(default: %(default)s)")
        self.parser.add_argument("--gbr-org",
                                 default=DEFAULT_GBR_ORG,
                                 metavar=META_NAME,
                                 help="Organisation name to include in GBR "
                                      "(default: %(default)s)")
        self.parser.add_argument("--gbr-email",
                                 default=DEFAULT_GBR_EMAIL,
                                 metavar=META_ADDR,
                                 help="Email address to include in GBR "
                                      "(default: %(default)s)")

    def run(self,
            parsed_args: Args,
            *args: typing.Any,
            **kwargs: typing.Any) -> Return:
        """Run with the given arguments."""
        log.info("setting up rpkimancer library objects")
        from ..cert import CertificateAuthority, TACertificateAuthority
        from ..sigobj import RouteOriginAttestation, RpkiGhostbusters
        # create CAs
        log.info("creating TA certificate authority")
        ta = TACertificateAuthority(as_resources=parsed_args.ta_as_resources,
                                    ip_resources=parsed_args.ta_ip_resources)
        log.info("creating suboridinate certificate authority")
        ca = CertificateAuthority(issuer=ta,
                                  as_resources=parsed_args.ca_as_resources,
                                  ip_resources=parsed_args.ca_ip_resources)
        # create ROA
        log.info("creating ROA object")
        RouteOriginAttestation(issuer=ca,
                               as_id=parsed_args.roa_asid,
                               ip_address_blocks=parsed_args.roa_networks)
        # create GBR
        log.info("creating ghostbusters record object")
        RpkiGhostbusters(issuer=ca,
                         full_name=parsed_args.gbr_full_name,
                         org=parsed_args.gbr_org,
                         email=parsed_args.gbr_email)
        # run plugins
        log.info("running plugins")
        plugin_publish_kwargs = dict()
        for plugin in self._plugins:
            log.info(f"running plugin {plugin}")
            if (kwargs := plugin(parsed_args, ca)) is not None:
                log.debug(f"{plugin} returned kwargs for publish: {kwargs}")
                plugin_publish_kwargs.update(kwargs)
        # publish objects
        log.info(f"publishing in-memory objects to {parsed_args.output_dir}")
        ta.publish(pub_path=os.path.join(parsed_args.output_dir, PUB_SUB_DIR),
                   tal_path=os.path.join(parsed_args.output_dir, TAL_SUB_DIR),
                   **plugin_publish_kwargs)
        return None


PluginReturn = typing.Optional[typing.Mapping[str, str]]


class ConjurePlugin:
    """Base class for conjure subcommand plugins."""

    def __init__(self, parent: argparse.ArgumentParser) -> None:
        """Initialise the plugin."""
        self.parser = parent
        self.init_parser()

    def __call__(self,
                 parsed_args: argparse.Namespace,
                 ca: CertificateAuthority,
                 *args: typing.Any,
                 **kwargs: typing.Any) -> PluginReturn:
        """Run the plugin."""
        return self.run(parsed_args, ca, *args, **kwargs)

    def init_parser(self) -> None:
        """Set up command line argument parser."""
        raise NotImplementedError

    def run(self,
            parsed_args: Args,
            ca: CertificateAuthority,
            *args: typing.Any,
            **kwargs: typing.Any) -> PluginReturn:
        """Run with the given arguments, returning extra publish kwargs."""
        raise NotImplementedError



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""rpkincant perceive command implementation."""

from __future__ import annotations

import contextlib
import functools
import logging
import os
import sys
import typing

from . import Args, BaseCommand, Return

if typing.TYPE_CHECKING:
    from rpkimancer.asn1 import Interface

log = logging.getLogger(__name__)

WriteGenerator = typing.Generator[typing.Callable[["Interface", str], None],
                                  None,
                                  None]


class Perceive(BaseCommand):
    """Learn of the secrets carried within an RPKI signed object."""

    subcommand = "perceive"
    usage = None

    def init_parser(self) -> None:
        """Set up command line argument parser."""
        self.parser.add_argument("paths", nargs="+",
                                 metavar="<file-path>",
                                 help="Path to the signed object")
        self.parser.add_argument("--output", "-o",
                                 metavar="<output-path>",
                                 type=self._output, default=self._output(),
                                 help="Path to output file "
                                      "(default: STDOUT)")
        info_group = self.parser.add_argument_group("output options")
        info = info_group.add_mutually_exclusive_group()
        self.parser.set_defaults(info_property="econtent")
        info.add_argument("--signed-data", "-S", dest="info_property",
                          action="store_const", const=None,
                          help="Print decoded SignedData")
        info.add_argument("--econtent-info", "-I", dest="info_property",
                          action="store_const", const="econtent_info",
                          help="Print decoded EncapsulatedContentInfo")
        info.add_argument("--econtent", "-E", dest="info_property",
                          action="store_const", const="econtent",
                          help="Print decoded eContent (default)")
        fmt_group = self.parser.add_argument_group("format options")
        fmt = fmt_group.add_mutually_exclusive_group()
        self.parser.set_defaults(fmt_method="to_txt")
        fmt.add_argument("--asn1", "-A", dest="fmt_method",
                         action="store_const", const="to_asn1",
                         help="Output ASN.1 data syntax")
        fmt.add_argument("--json", "-j", dest="fmt_method",
                         action="store_const", const="to_json",
                         help="Output JSON encoded data")
        fmt.add_argument("--jer", "-J", dest="fmt_method",
                         action="store_const", const="to_jer",
                         help="Output JER encoded data")
        fmt.add_argument("--raw", "-R", dest="fmt_method",
                         action="store_const", const="to_internal",
                         help="Output internal python data representation")

    def run(self,
            parsed_args: Args,
            *args: typing.Any,
            **kwargs: typing.Any) -> Return:
        """Run with the given arguments."""
        log.info("setting up rpkimancer library objects")
        from ..sigobj import from_ext
        objects = list()
        for path in parsed_args.paths:
            log.info(f"deciphering {path}")
            _, ext = os.path.splitext(path)
            if not ext:
                continue
            log.info(f"trying to get object type for file extension {ext}")
            try:
                object_cls = from_ext(ext)
            except KeyError:
                log.warning(f"no signed object type with file extension {ext}")
                continue
            log.info(f"trying to read signed object from {path}")
            with open(path, "rb") as f:
                data = f.read()
            log.info(f"trying to deserialise to {object_cls}")
            obj = object_cls.from_der(data)
            objects.append(obj)
        with parsed_args.output as write:
            for obj in objects:
                write(obj, parsed_args.info_property, parsed_args.fmt_method)
        return None

    @staticmethod
    @contextlib.contextmanager
    def _output(path: typing.Optional[str] = None) -> WriteGenerator:

        def _write(obj: Interface,
                   info_property: typing.Optional[str],
                   fmt_method: str,
                   f: typing.TextIO) -> None:
            if info_property is not None:
                obj = getattr(obj, info_property)
            func = getattr(obj, fmt_method)
            f.write(f"{func()}\n")

        if path is None:
            yield functools.partial(_write, f=sys.stdout)
        else:
            with open(path, "w") as f:
                yield functools.partial(_write, f=f)



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""RPKI Certificate Authority implementation - RFC6487."""

from __future__ import annotations

import base64
import datetime
import logging
import os
import typing

from cryptography import x509
from cryptography.hazmat.primitives import serialization

from . import base, oid

if typing.TYPE_CHECKING:
    from ..sigobj import RpkiManifest

log = logging.getLogger(__name__)


class CertificateAuthority(base.BaseResourceCertificate):
    """RPKI Certificate Authority - RFC6487."""

    def __init__(self, *,
                 common_name: str = "CA",
                 crl_days: int = 7,
                 mft_days: int = 7,
                 **kwargs: typing.Any) -> None:
        """Initialise the Certificate Authority."""
        log.info(f"doing initialisation of {self} as CertificateAuthority")
        self._issued: base.ResourceCertificateList = list()
        self.next_serial_number = 1
        super().__init__(common_name=common_name, ca=True, **kwargs)
        # rfc 6487 section 5
        self.crl_days = crl_days
        self.next_crl_number = 0
        self._crl: typing.Optional[x509.CertificateRevocationList] = None
        self.issue_crl()
        self.mft_days = mft_days
        self.next_mft_number = 0

    @property
    def crl(self) -> typing.Optional[x509.CertificateRevocationList]:
        """Get the last CRL issued by this CA."""
        return self._crl

    @property
    def crl_der(self) -> bytes:
        """Get the last CRL as DER-encoded bytes."""
        return typing.cast(x509.CertificateRevocationList,
                           self.crl).public_bytes(serialization.Encoding.DER)

    @property
    def repo_path(self) -> str:
        """Get the filesystem path to this CA's publication point."""
        return os.path.join(typing.cast(CertificateAuthority,
                                        self.issuer).repo_path,
                            self.subject_cn)

    @property
    def cert_path(self) -> str:
        """Get the filesystem path to cert in the issuer publication point."""
        return os.path.join(typing.cast(CertificateAuthority,
                                        self.issuer).repo_path,
                            f"{self.subject_cn}.cer")

    @property
    def mft_entry(self) -> typing.Optional[base.ManifestEntryInfo]:
        """Get an entry for inclusion in the issuer's manifest."""
        return (os.path.basename(self.cert_path), self.cert_der)

    @property
    def crl_path(self) -> str:
        """Get the filesystem path to the CRL in publication point."""
        return os.path.join(self.repo_path, "revoked.crl")

    @property
    def mft_path(self) -> str:
        """Get the filesystem path to the manifest in publication point."""
        return os.path.join(self.repo_path, "manifest.mft")

    @property
    def issued(self) -> base.ResourceCertificates:
        """Get a generator over the certifactes issued by this CA."""
        for cert in self._issued:
            yield cert

    @property
    def crldp(self) -> typing.Optional[x509.CRLDistributionPoints]:
        """Get the CRLDistributionPoint extension for the certificate."""
        crldp_uri = f"{self.base_uri}/{self.crl_path}"
        crldp = x509.CRLDistributionPoints([
            x509.DistributionPoint([x509.UniformResourceIdentifier(crldp_uri)],
                                   relative_name=None,
                                   reasons=None,
                                   crl_issuer=None),
        ])
        return crldp

    @property
    def aia(self) -> typing.Optional[x509.AuthorityInformationAccess]:
        """Get the AuthorityInformationAccess extension for the certificate."""
        aia_uri = f"{self.base_uri}/{self.cert_path}"
        aia = x509.AuthorityInformationAccess([
            x509.AccessDescription(oid.AIA_CA_ISSUERS_OID,
                                   x509.UniformResourceIdentifier(aia_uri)),
        ])
        return aia

    @property
    def sia(self) -> typing.Optional[x509.SubjectInformationAccess]:
        """Get the SubjectInformationAccess extension for the certificate."""
        sia_repo_uri = f"{self.base_uri}/{self.repo_path}"
        sia_mft_uri = f"{self.base_uri}/{self.mft_path}"
        sia = x509.SubjectInformationAccess([
            x509.AccessDescription(oid.SIA_CA_REPOSITORY_OID,
                                   x509.UniformResourceIdentifier(sia_repo_uri)),  # noqa: E501
            x509.AccessDescription(oid.SIA_MFT_ACCESS_OID,
                                   x509.UniformResourceIdentifier(sia_mft_uri)),  # noqa: E501
        ])
        return sia

    def issue_cert(self,
                   subject: typing.Optional[base.BaseResourceCertificate] = None) -> x509.Certificate:  # noqa: E501
        """Issue a new Resource Certificate with this CA."""
        if subject is None:
            subject = self
        cert = subject.cert_builder.sign(private_key=self.private_key,
                                         algorithm=self.hash_algorithm())
        self._issued.append(subject)
        self.next_serial_number += 1
        return cert

    def issue_crl(self,
                  to_revoke: typing.Optional[base.ResourceCertificates] = None) -> None:  # noqa: E501
        """Issue a new CRL for this CA."""
        now = datetime.datetime.utcnow()
        next_update = now + datetime.timedelta(days=self.crl_days)
        crl_builder = x509.CertificateRevocationListBuilder()
        crl_builder = crl_builder.issuer_name(self.cert.subject)
        crl_builder = crl_builder.last_update(now)
        crl_builder = crl_builder.next_update(next_update)
        aki = x509.AuthorityKeyIdentifier\
                  .from_issuer_public_key(self.public_key)
        crl_builder = crl_builder.add_extension(aki, critical=False)
        crl_number = x509.CRLNumber(self.next_crl_number)
        crl_builder = crl_builder.add_extension(crl_number, critical=False)
        if self.crl is not None:
            for revoked in self.crl:
                # TODO: clean up expired certs
                crl_builder = crl_builder.add_revoked_certificate(revoked)
        if to_revoke is not None:
            for c in to_revoke:
                rc_builder = x509.RevokedCertificateBuilder()
                rc_builder = rc_builder.revocation_date(now)
                rc_builder = rc_builder.serial_number(c.cert.serial_number)
                revoked_cert = rc_builder.build()
                crl_builder = crl_builder.add_revoked_certificate(revoked_cert)
        self._crl = crl_builder.sign(self.private_key, self.hash_algorithm())
        self.next_crl_number += 1

    def issue_mft(self,
                  file_list: typing.List[base.ManifestEntryInfo]) -> None:
        """Issue a new manifest for this CA."""
        now = datetime.datetime.utcnow()
        next_update = now + datetime.timedelta(days=self.mft_days)
        from ..sigobj import RpkiManifest
        self._mft = RpkiManifest(issuer=self,
                                 file_name=os.path.basename(self.mft_path),
                                 manifest_number=self.next_mft_number,
                                 this_update=now,
                                 next_update=next_update,
                                 file_list=file_list)
        self.next_mft_number += 1

    @property
    def mft(self) -> RpkiManifest:
        """Get the last manifest issued by this CA."""
        return self._mft

    def publish(self, *,
                pub_path: str,
                recursive: bool = True,
                **kwargs: typing.Any) -> None:
        """Publish this CA's artifacts as DER files in the PP."""
        mft_file_list = list()
        full_pub_path = os.path.join(pub_path, self.uri_path)
        os.makedirs(os.path.join(full_pub_path, self.repo_path), exist_ok=True)
        with open(os.path.join(full_pub_path, self.cert_path), "wb") as f:
            f.write(self.cert_der)
        with open(os.path.join(full_pub_path, self.crl_path), "wb") as f:
            f.write(self.crl_der)
        mft_file_list.append((os.path.basename(self.crl_path),
                              self.crl_der))
        for issuee in self.issued:
            if issuee is not self:
                if issuee.mft_entry is not None:
                    mft_file_list.append(issuee.mft_entry)
                if recursive is True:
                    issuee.publish(pub_path=pub_path,
                                   recursive=recursive,
                                   **kwargs)
        self.issue_mft(mft_file_list)
        with open(os.path.join(full_pub_path, self.mft_path), "wb") as f:
            f.write(self.mft.to_der())


class TACertificateAuthority(CertificateAuthority):
    """RPKI Trust Anchor Certificate Authority - RFC6487."""

    def __init__(self, *,
                 common_name: str = "TA",
                 base_uri: str = "rsync://rpki.example.net/rpki",
                 **kwargs: typing.Any) -> None:
        """Initialise the Certificate Authority."""
        log.info(f"doing initialisation of {self} as TACertificateAuthority")
        super().__init__(common_name=common_name, issuer=None, **kwargs)

    @property
    def repo_path(self) -> str:
        """Get the filesystem path to this CA's publication point."""
        return self.subject_cn

    @property
    def cert_path(self) -> str:
        """Get the filesystem path to cert in the publication point root."""
        return f"{self.subject_cn}.cer"

    @property
    def tal_path(self) -> str:
        """Get the filesystem path to the trust anchor locator."""
        return f"{self.subject_cn}.tal"

    @property
    def tal(self) -> bytes:
        """Get the contents of the TAL for this trust anchor."""
        tal_contents = f"{self.base_uri}/{self.cert_path}\n\n".encode()
        tal_contents += base64.b64encode(self.subject_public_key_info.to_der())
        return tal_contents

    def publish(self, *,
                pub_path: str,
                tal_path: typing.Optional[str] = None,
                recursive: bool = True,
                **kwargs: typing.Any) -> None:
        """Publish this CA's artifacts and TAL."""
        super().publish(pub_path=pub_path, recursive=recursive, **kwargs)
        if tal_path is not None:
            os.makedirs(tal_path, exist_ok=True)
            with open(os.path.join(tal_path, self.tal_path), "wb") as f:
                f.write(self.tal)



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Modules implementing CLI tools using rpkimancer."""

from __future__ import annotations

import argparse
import logging
import sys
import typing

import argcomplete

log = logging.getLogger(__name__)

OptionalSubParser = typing.Optional[argparse._SubParsersAction]
Args = argparse.Namespace
OptionalArgv = typing.Optional[typing.List[str]]
OptionalArgs = typing.Union[Args, OptionalArgv]
Return = typing.Optional[int]


def set_log_level(verbosity: int) -> None:
    """Set logging verbosity."""
    level = logging.WARNING - (10 * verbosity)
    logging.basicConfig(level=level)


class BaseCommand:
    """Base class for command-line utilities."""

    subcommand: str
    usage: typing.Optional[str] = "%(prog)s [options]"

    def __init__(self, parent: OptionalSubParser = None) -> None:
        """Initialise the command."""
        description = self.__doc__
        help_fmt = argparse.RawTextHelpFormatter
        if parent is None:
            parser = argparse.ArgumentParser(description=description,
                                             usage=self.usage,
                                             formatter_class=help_fmt)
        else:
            parser = parent.add_parser(self.subcommand,
                                       description=description,
                                       help=description,
                                       formatter_class=help_fmt)
            parser.set_defaults(run=self)
        parser.add_argument("-v", dest="verbosity",
                            action="count", default=0,
                            help="Increase logging verbosity:\n"
                                 "-v: INFO level logging\n"
                                 "-vv: DEBUG level logging\n"
                                 "-vvv: DEBUG logging and tracebacks")
        self.parser = parser
        self.init_parser()

    def __call__(self,
                 args_or_argv: OptionalArgs = None,
                 *args: typing.Any,
                 **kwargs: typing.Any) -> Return:
        """Run the command."""
        if isinstance(args_or_argv, Args):
            parsed_args = args_or_argv
        else:
            argcomplete.autocomplete(self.parser,
                                     always_complete_options="long")
            if args_or_argv is None:  # pragma: no cover
                argv = sys.argv[1:]
            else:
                argv = args_or_argv
            parsed_args = self.parser.parse_args(argv)
            set_log_level(parsed_args.verbosity)
        try:
            return self.run(parsed_args, *args, **kwargs)
        except KeyboardInterrupt:  # pragma: no cover
            log.error("Interrupted by Ctrl+C")
            return 2
        except Exception as e:  # pragma: no cover
            log.error(f"{e!r}", exc_info=(parsed_args.verbosity >= 3))
            return 1

    def init_parser(self) -> None:
        """Set up command line argument parser."""
        raise NotImplementedError

    def run(self,
            parsed_args: Args,
            *args: typing.Any,
            **kwargs: typing.Any) -> Return:
        """Run with the given arguments."""
        raise NotImplementedError



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""ASN.1 data types and helpers."""
from __future__ import annotations

import contextlib
import logging
import threading
import typing

import pycrate_asn1rt.init

from .types import ASN1Class, ASN1Obj, ASN1ObjData
from ..utils import LogWriter

log = logging.getLogger(__name__)

log_writer = LogWriter(log, level=logging.WARNING)

InterfaceSubclass = typing.TypeVar("InterfaceSubclass",
                                   bound="Interface")


def append_info_object_set(obj_set: ASN1Class, *obj_ins: ASN1Class) -> None:
    """Append an instance to an existing object information set at runtime."""
    for ins in obj_ins:
        log.info(f"Adding {obj_ins} to constraining object info set {obj_set}")
        obj_set.get_val().root.append(ins.get_val())
    log.info(f"re-building lookup table for {obj_set}")
    pycrate_asn1rt.init.build_classset_dict(obj_set)


class Interface:
    """Generic base ASN.1 type wrapping pycrates API."""

    content_syntax: ASN1Obj
    _lock = threading.Lock()

    @classmethod
    def __init_subclass__(cls, /, **kwargs: typing.Any) -> None:
        """Create a new lock for each subclass."""
        super().__init_subclass__(**kwargs)
        cls._lock = threading.Lock()

    def __init__(self, data: typing.Any) -> None:
        """Initialise the instance from python data."""
        log.info(f"starting initialisation of {self} ASN.1 content")
        with self.constructed(data) as instance:
            self._content_data = instance.get_val()
        log.info(f"finished initialisation of {self} ASN.1 content")

    @classmethod
    def from_data(cls: typing.Type[InterfaceSubclass],
                  data: typing.Any) -> InterfaceSubclass:
        """Construct an instance from python data."""
        log.info(f"creating new {cls} object")
        self: InterfaceSubclass = cls.__new__(cls)
        Interface.__init__(self, data)
        return self

    @classmethod
    def from_der(cls: typing.Type[InterfaceSubclass],
                 der_data: bytes) -> InterfaceSubclass:
        """Construct an instance from DER encoded data."""
        log.info(f"trying to acquire lock for {cls}")
        with cls._lock:
            log.info(f"deserialising {cls} object from DER data.")
            with log_writer.redirect_stdout():
                cls.content_syntax.from_der(der_data)
            data = cls.content_syntax.get_val()
            cls.content_syntax.reset_val()
            log.info(f"finished deserialising {cls} object")
        self: InterfaceSubclass = cls.__new__(cls)
        Interface.__init__(self, data)
        return self

    @property
    def content_data(self) -> ASN1ObjData:
        """Get the underlying python data for this type instance."""
        return self._content_data

    @contextlib.contextmanager
    def constructed(self,
                    data: typing.Optional[ASN1ObjData] = None) -> ASN1Obj:
        """Provide a context manager to mediate the global pycrates object."""
        if data is None:
            data = self.content_data
        log.info(f"trying to acquire lock for {self.__class__}")
        with self._lock:
            log.debug(f"instantiating ASN1Obj from data: {data}")
            try:
                self.content_syntax.set_val(data)
                yield self.content_syntax
            finally:
                self.content_syntax.reset_val()

    def to_txt(self) -> str:
        """Get default text serialization."""
        return self.to_asn1()

    def to_asn1(self) -> str:
        """Serialize as ASN.1 data."""
        with self.constructed() as instance:
            log.info(f"serialising object {self} to ASN.1 data encoding")
            with log_writer.redirect_stdout():
                val = instance.to_asn1()
            log.info(f"finished serialising object {self}")
        return typing.cast(str, val)

    def to_der(self) -> bytes:
        """Serialize as DER."""
        with self.constructed() as instance:
            log.info(f"serialising object {self} to DER")
            with log_writer.redirect_stdout():
                val = instance.to_der()
            log.info(f"finished serialising object {self}")
        return typing.cast(bytes, val)

    def to_jer(self) -> str:
        """Serialize as JER."""
        with self.constructed() as instance:
            log.info(f"serialising object {self} to JSON")
            with log_writer.redirect_stdout():
                val = instance.to_jer()
            log.info(f"finished serialising object {self}")
        return typing.cast(str, val)

    def to_json(self) -> str:
        """Serialize as JSON."""
        return self.to_jer()

    def to_internal(self) -> str:
        """Serialize using internal python repr."""
        return f"{self.content_data!r}"



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Number Resource Extension implementations - RFC3779."""
from __future__ import annotations

import logging
import typing

from cryptography import x509

from . import asn1, oid
from ..asn1.mod import IPAddrAndASCertExtn
from ..asn1.types import ASN1Class
from ..resources import (ASIdentifiers, AsResourcesInfo,
                         IPAddrBlocks, IpResourcesInfo)

log = logging.getLogger(__name__)


class X509CertificateExtension(x509.UnrecognizedExtension):
    """Custom certificate extension with ASN.1 handling."""

    @classmethod
    def __init_subclass__(cls,
                          ext_type: typing.Optional[ASN1Class] = None,
                          **kwargs: typing.Any) -> None:
        """Register the EXTENSION instance for DER encoding/decoding."""
        super().__init_subclass__(**kwargs)
        if ext_type is not None:
            asn1.Certificate.register_ext_type(ext_type)


class IpResources(X509CertificateExtension,
                  ext_type=IPAddrAndASCertExtn.ext_IPAddrBlocks):
    """IP Address Resources X.509 certificate extension - RFC3779."""

    def __init__(self, ip_resources: IpResourcesInfo) -> None:
        """Initialise the certificate extension."""
        ip_address_blocks_data = IPAddrBlocks(ip_resources).to_der()
        super().__init__(oid.IP_RESOURCES_OID, ip_address_blocks_data)


class AsResources(X509CertificateExtension,
                  ext_type=IPAddrAndASCertExtn.ext_ASIdentifiers):
    """AS Number Resources X.509 certificate extension - RFC3779."""

    def __init__(self, as_resources: AsResourcesInfo) -> None:
        """Initialise the certificate extension."""
        as_identifiers_data = ASIdentifiers(as_resources).to_der()
        super().__init__(oid.AS_RESOURCES_OID, as_identifiers_data)



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""RPKI Ghostbusters Record implementation - RFC6493."""

from __future__ import annotations

import json
import logging
import typing

from .base import EncapsulatedContentType, SignedObject
from ..asn1.mod import RPKIGhostbusters
from ..resources import INHERIT_AS, INHERIT_IPV4, INHERIT_IPV6

log = logging.getLogger(__name__)


class RpkiGhostbustersContentType(EncapsulatedContentType):
    """encapContentInfo for RPKI Ghostbusters Record - RFC6493."""

    asn1_definition = RPKIGhostbusters.ct_rpkiGhostbusters
    file_ext = "gbr"
    as_resources = INHERIT_AS
    ip_resources = [INHERIT_IPV4, INHERIT_IPV6]

    def __init__(self,
                 full_name: str,
                 org: typing.Optional[str] = None,
                 address: typing.Optional[str] = None,
                 tel: typing.Optional[str] = None,
                 email: typing.Optional[str] = None) -> None:
        """Initialise the encapContentInfo."""
        log.info(f"preparing data for {self}")
        vcard = "BEGIN:VCARD\r\n"
        vcard += "VERSION:4.0\r\n"
        vcard += f"FN:{full_name}\r\n"
        if org is not None:
            vcard += f"ORG:{org}\r\n"
        if address is not None:
            vcard += f"ADR:{address}\r\n"
        if tel is not None:
            vcard += f"TEL;VALUE=uri:tel:{tel}\r\n"
        if email is not None:
            vcard += f"EMAIL:{email}\r\n"
        vcard += "END:VCARD"
        data = vcard.encode()
        super().__init__(data)

    def to_txt(self) -> str:
        """Get default text serialization."""
        return typing.cast(str, self.content_data.decode())

    def to_json(self) -> str:
        """Serialize as JSON."""
        vcard = self.to_txt()
        data = dict()
        for line in vcard.splitlines():
            key, val = line.split(":", 1)
            if key in ("BEGIN", "END"):
                continue
            data[key.lower()] = val
        return json.dumps(data, indent=2)


class RpkiGhostbusters(SignedObject[RpkiGhostbustersContentType]):
    """CMS ASN.1 ContentInfo for RPKI Ghostbusters Records."""



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Message digest algoritms used in rpkimancer library."""

from __future__ import annotations

import hashlib
import logging
import typing

log = logging.getLogger(__name__)


DigestAlgorithm = typing.Callable[[bytes], "hashlib._Hash"]
AlgorithmDict = typing.Dict[typing.Tuple[int, ...], DigestAlgorithm]

SHA256: typing.Final = (2, 16, 840, 1, 101, 3, 4, 2, 1)

DIGEST_ALGORITHMS: typing.Final[AlgorithmDict] = {SHA256: hashlib.sha256}



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""rpkimancer CLI tests package."""

from __future__ import annotations

import copy
import importlib.abc
import importlib.metadata
import ipaddress
import logging
import os
import subprocess
import sys

import pytest

log = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def target_directory(tmp_path_factory):
    """Set up a tmp directory for writing artifacts."""
    return tmp_path_factory.mktemp("target")


@pytest.fixture(scope="session")
def patch_meta_path():
    """Inject a dummy plugin distribution into 'meta_path'."""

    class DummyDistribution(importlib.metadata.Distribution):
        def read_text(self, filename):
            if filename == "PKG-INFO":
                text = ["Metadata-Version: 2.1",
                        "Name: rpkimancer-foo",
                        "Version: 0.0.1"]
            elif filename == "entry_points.txt":
                text = ["[rpkimancer.asn1.modules]",
                        "RpkiFoo = rpkimancer_foo.asn1",
                        "[rpkimancer.cli.conjure]",
                        "ConjureFoo = rpkimancer_foo.conjure:ConjureFoo",
                        "[rpkimancer.sigobj]",
                        "FooObject = rpkimancer_foo.sigobj:FooObject"]
            else:
                return ""
            return "\n".join(text)

        def locate_file(self, path):
            raise NotImplementedError

    dummy_finder_ctx = importlib.metadata.DistributionFinder.Context()

    class DummyMetaPathFinder(importlib.abc.MetaPathFinder):
        def find_spec(self, fullname, path, target=None):
            return None

        def find_distributions(self, context=dummy_finder_ctx):
            yield DummyDistribution()

    try:
        old_meta_path = copy.copy(sys.meta_path)
        sys.meta_path.append(DummyMetaPathFinder)
        yield
    finally:
        sys.meta_path = old_meta_path


@pytest.mark.usefixtures("patch_meta_path")
class TestCli:
    """Test cases for rpkimancer CLI tools."""

    def test_conjure(self, target_directory):
        """Test the conjure subcommand."""
        from rpkimancer.cli.__main__ import main
        argv = ["conjure", "--output-dir", f"{target_directory}"]
        retval = main(argv)
        assert retval is None

    @pytest.mark.parametrize("out", (None, "-E", "-I", "-S"))
    @pytest.mark.parametrize("fmt", (None, "-A", "-j", "-J", "-R"))
    @pytest.mark.parametrize("ext", ("gbr", "mft", "roa"))
    def test_perceive(self, target_directory, ext, fmt, out):
        """Test the perceive subcommand."""
        from rpkimancer.cli.__main__ import main
        repo_path = target_directory / "repo"
        paths = repo_path.rglob(f"*.{ext}")
        argv = ["perceive",
                "--output", os.devnull]
        if fmt is not None:
            argv.append(fmt)
        if out is not None:
            argv.append(out)
        argv.extend(str(p) for p in paths)
        retval = main(argv)
        assert retval is None

    @pytest.mark.rpki_client
    @pytest.mark.parametrize("iteration", range(10))
    def test_rpki_validate(self, tmp_path_factory, iteration):
        """Test rpki-client can validate the generated artifacts."""
        target_directory = tmp_path_factory.mktemp("target")
        from rpkimancer.cli.__main__ import main
        argv = ["conjure", "--output-dir", f"{target_directory}"]
        retval = main(argv)
        assert retval is None
        cmd = ["rpki-client", "-jnvv"]
        repo_path = target_directory / "repo"
        cmd.extend(("-d", str(repo_path)))
        tal_paths = (target_directory / "tals").glob("*.tal")
        for tal_path in tal_paths:
            cmd.extend(("-t", str(tal_path)))
        output_path = target_directory / "output"
        output_path.mkdir(exist_ok=True)
        cmd.append(str(output_path))
        proc = subprocess.run(cmd,
                              text=True,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.STDOUT)
        output = proc.stdout
        valid = True
        for line in output.rstrip().splitlines():
            if "signature failure" in line:
                level = logging.ERROR
                valid = False
            else:
                level = logging.INFO
            log.log(level, line.strip())
        assert proc.returncode == 0
        assert valid


class TestHelpers:
    """Test cases for cli argument helpers."""

    @pytest.mark.parametrize(("input_str", "value"),
                             (("10.0.0.0/8", ipaddress.IPv4Network("10.0.0.0/8")),  # noqa: E501
                              ("2001:db8::/32", ipaddress.IPv6Network("2001:db8::/32")),  # noqa: E501
                              ("192.168.1.128-192.168.2.255", (ipaddress.IPv4Address("192.168.1.128"),  # noqa: E501
                                                               ipaddress.IPv4Address("192.168.2.255"))),  # noqa: E501
                              ("2001:db8:beef::-2001:db8:dead::", (ipaddress.IPv6Address("2001:db8:beef::"),  # noqa: E501
                                                                   ipaddress.IPv6Address("2001:db8:dead::")))))  # noqa: E501
    def test_ip_resource_helper(self, input_str, value):
        """Test the 'ip_resource' arg type helper."""
        from rpkimancer.cli.helpers import ip_resource
        assert ip_resource(input_str) == value

    @pytest.mark.parametrize(("input_str", "value"),
                             (("10.0.0.0/8", (ipaddress.IPv4Network("10.0.0.0/8"), None)),  # noqa: E501
                              ("2001:db8::/32", (ipaddress.IPv6Network("2001:db8::/32"), None)),  # noqa: E501
                              ("192.168.0.0/16-24", (ipaddress.IPv4Network("192.168.0.0/16"), 24)),  # noqa: E501
                              ("2001:db8:f00::/48-64", (ipaddress.IPv6Network("2001:db8:f00::/48"), 64))))  # noqa: E501
    def test_roa_network_helper(self, input_str, value):
        """Test the 'roa_network' arg type helper."""
        from rpkimancer.cli.helpers import roa_network
        assert roa_network(input_str) == value

    @pytest.mark.parametrize(("input_str", "value"),
                             (("65000", 65000),
                              ("65001-65005", (65001, 65005))))
    def test_as_id_or_range_helper(self, input_str, value):
        """Test the 'as_id_or_range' arg type helper."""
        from rpkimancer.cli.helpers import as_id_or_range
        assert as_id_or_range(input_str) == value



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Compile and re-export the provided ASN.1 modules."""
from __future__ import annotations

import logging
import typing

import pycrate_asn1rt.asnobj as _asn1_object_types
import pycrate_asn1rt.asnobj_basic as _asn1_basic_types
import pycrate_asn1rt.asnobj_class as _asn1_class_types

log = logging.getLogger(__name__)

ASN1Obj = _asn1_object_types.ASN1Obj
ASN1ObjData = typing.Any
OID = _asn1_basic_types.OID
ASN1Class = _asn1_class_types.CLASS



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""rpkincant cli argument helpers."""

from __future__ import annotations

import ipaddress
import typing

if typing.TYPE_CHECKING:
    from ..resources import IPAddressFamilyInfo, ASIdOrRangeInfo
    from ..sigobj.roa import RoaNetworkInfo


def ip_resource(input_str: str) -> IPAddressFamilyInfo:
    """Convert input string to IPAddressFamilyInfo variant."""
    try:
        return ipaddress.ip_network(input_str)
    except ValueError:
        lower, upper = input_str.split("-", 1)
    try:
        return ipaddress.IPv4Address(lower), ipaddress.IPv4Address(upper)
    except ValueError:
        return ipaddress.IPv6Address(lower), ipaddress.IPv6Address(upper)


def roa_network(input_str: str) -> RoaNetworkInfo:
    """Convert input string to RoaNetworkInfo tuple."""
    try:
        network, maxlen = input_str.split("-", 1)
        return (ipaddress.ip_network(network), int(maxlen))
    except ValueError:
        return (ipaddress.ip_network(input_str), None)


def as_id_or_range(input_str: str) -> ASIdOrRangeInfo:
    """Convert input string to ASIdOrRangeInfo variant."""
    try:
        return int(input_str)
    except ValueError:
        lower, upper = input_str.split("-", 1)
        return int(lower), int(upper)
